/*
 *  client.c
 *  Echo Client
 *
 *  Created by Ben Tribelhorn on 4/29/06.
 *
 */

#include <strings.h> 
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>

//(btribelh)
#define UID 8000
#define MAXLINE 80

ssize_t Write(int fd, void *buf, size_t n)
{
	size_t nleft = n;
	ssize_t nwritten;
	char *bufp = buf;
	
	while(nleft > 0)
	{
		if((nwritten = write(fd, bufp, nleft)) <= 0)
			//if(errno == EINTR)	//interrupted by sig handler
			//	nwritten = 0;	//try again
			//else
				return -1;		//other error
		nleft -= nwritten;
		bufp += nwritten;
	}
	
	return n;
}

ssize_t Readline(int fd, void *buf, size_t maxlen)
{
	int n, cnt;
	char c, *bufp = buf;
	
	for(n = 1; n<maxlen; n++)
	{
		if((cnt = read(fd, &c, 1)) == 1)
		{
			*bufp++ = c;
			if(c == '\n')
				break;
		}
		else if(cnt == 0)
		{
			if(n == 1)
				return 0;	//EOF, no data
			else
				break;		//EOF, some data was read
		}
		else
			return -1;		//error
	}
	*bufp = 0;
	return n;
}

int main(int argc, char **argv)
{
	int clientfd, port;
	char *host, buf[MAXLINE];
	
	host = argv[1];
	port = UID;//atoi(argv[2]);
	
	//-------------------------------------
	struct hostent *hp;
	struct sockaddr_in serveraddr;
	
	if((clientfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		return -1;
		
	//Get IP address and port
	if((hp = gethostbyname(host)) == NULL)
		return -2;
	bzero((void *) &serveraddr, sizeof(serveraddr));
	serveraddr.sin_family = AF_INET;
	bcopy((char *)*(hp->h_addr_list), (char *)&serveraddr.sin_addr.s_addr, hp->h_length);
	serveraddr.sin_port = htons(port);
	
	if(connect(clientfd, (struct sockaddr *) &serveraddr, sizeof(serveraddr)) < 0)
		return -1;
	//-------------------------------------
	
	fputs("Enter msg: ", stdout);
	while(fgets(buf, MAXLINE, stdin) != NULL)
	{
		Write(clientfd, buf, strlen(buf));
		Readline(clientfd, buf, MAXLINE);
		fputs(buf, stdout);
		fputs("Enter msg: ", stdout);
	}
	
	close(clientfd);

	return 0;
}
