/*
 *  server.c
 *  Echo Server
 *
 *  Created by Ben Tribelhorn on 4/29/06.
 *  Last Modified 6/20/06.
 *
 */
 
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <strings.h>

#include "vdigGrab.h"
#include "definitions.h"

#define PORT	8000
#define LISTENQ 0

ssize_t Readline(int fd, void *buf, size_t maxlen);
ssize_t Write(int fd, void *buf, size_t n);

void getInput(int connfd)
{
	size_t n;
	char buf[MAXLINE];
	
	while((n = Readline(connfd, buf, MAXLINE)) != 0)
	{
	  if(n == -1) //In some cases this happens when the client hits ctr-c
		{
			printf("Connection unexpectedly closed.\n");
			return;
		}
		else
		{
			//printf("server received %d bytes.\n", (int)n);
			
			//echo back what we got for testing
			//Write(connfd, buf, n);
			
			//do something with the data here
			int new_n = n;	//Avoids some stupid compiler warning/err
			vdgDataProcess(buf, &new_n);
			Write(connfd, buf, (size_t)new_n);
		}
	}
}

ssize_t Write(int fd, void *buf, size_t n)
{
	size_t nleft = n;
	ssize_t nwritten;
	char *bufp = buf;
	
	//write as many bytes as you can until done
	while(nleft > 0)
	{
		if((nwritten = write(fd, bufp, nleft)) <= 0)
			//if(errno == EINTR)	//interrupted by sig handler
			//	nwritten = 0;	//try again
			//else
				return -1;		//other error
		nleft -= nwritten;
		bufp += nwritten;
	}
	
	return n;
}

ssize_t Readline(int fd, void *buf, size_t maxlen)
{
	int n, cnt;
	char c, *bufp = buf;
	
	//read one byte at a time until up to the max length (or return)
	for(n = 1; n<maxlen; n++)
	{
		if((cnt = read(fd, &c, 1)) == 1)
		{
		        *bufp++ = c;   //Add char to buffer (byte)
			if(c == '\n')  //Done if they hit return
				break;
		}
		else if(cnt == 0)
		{
			if(n == 1)
				return 0;	//EOF, no data
			else
				break;		//EOF, some data was read
		}
		else
			return -1;		//error
	}
	*bufp = 0;
	return n;
}

int serverStart()//(int argc, char** argv)
{
	int listenfd, connfd, port, clientlen;
	struct sockaddr_in clientaddr;
	struct hostent *hp;
	char *haddrp;
	
	port = PORT;//atoi(argv[1]);
	
	//Create and configure the listening socket:
	//-------------------------------------
	int optval = 1;
	struct sockaddr_in serveraddr;
	
	//create descriptor
	if((listenfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		return -1;
	
	//elminate addy in use err from bind()
	if(setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, (const void *)&optval, sizeof(int)) < 0)
		return -1;
	
	//set up my port for listening on this host
	bzero((void *)&serveraddr, sizeof(serveraddr));
	serveraddr.sin_family = AF_INET;
	serveraddr.sin_addr.s_addr = htonl(INADDR_ANY);
	serveraddr.sin_port = htons((unsigned short)port);
	if(bind(listenfd, (struct sockaddr *)&serveraddr, sizeof(serveraddr)) < 0)
		return -1;
	
	//listen
	if(listen(listenfd, LISTENQ) < 0)
		return -1;
	//-------------------------------------
	
	while(1)
	{
		printf("Waiting for connection...\n");
		clientlen = sizeof(clientaddr);
		connfd = accept(listenfd, (struct sockaddr *)&clientaddr, &clientlen);
		hp = gethostbyaddr((const char*)&clientaddr.sin_addr.s_addr,
							sizeof(clientaddr.sin_addr.s_addr), AF_INET);
		haddrp = inet_ntoa(clientaddr.sin_addr);
		if(hp)
			printf("Server connected to %s (%s)\n", hp->h_name, haddrp);
		getInput(connfd);
		close(connfd);
	}

	return 0;
}
