/*
 *  vdigGrab.c
 *  seeSaw
 *
 *  Created by Daniel Heckenberg.
 *  Copyright (c) 2004 Daniel Heckenberg. All rights reserved.
 *  (danielh.seeSaw<at>cse<dot>unsw<dot>edu<dot>au)
 *  
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the right to use, copy, modify, merge, publish, communicate, sublicence, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included 
 * in all copies or substantial portions of the Software.
 * 
 * TO THE EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED 
 * "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT 
 * NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NON-INFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */
 
#include <GLUT/glut.h>

#include "SSUtility.h"
#include "vdigGrab.h"
#include "myMath.h"
#include "list.h"
#include <stdio.h>

#pragma mark ---- VdigGrab data ----

struct tagVdigGrab
{
	// State
	int					isPreflighted;
	int					isGrabbing;
	int					isRecording;

	// QT Components
	SeqGrabComponent	seqGrab; 
	SGChannel			sgchanVideo;
	ComponentInstance   vdCompInst;

	// Device settings
	ImageDescriptionHandle	vdImageDesc;
	Rect					vdDigitizerRect;		

	// Destination Settings
	CGrafPtr				dstPort;
	GWorldPtr				dstGWorld;
	ImageSequence			dstImageSeq;

	// Compression settings
	short				cpDepth;
	CompressorComponent cpCompressor;
	CodecQ				cpSpatialQuality;
	CodecQ				cpTemporalQuality;
	long				cpKeyFrameRate;
	Fixed				cpFrameRate;

};

#pragma mark ---- Vision Global Data ----

//A copy for making it easier on us
VdigGrab* pvdg;

//Color Spec
Rect box;
struct colorSpec
{
	int ymin, umin, vmin;
	int ymax, umax, vmax;
	int rmin, gmin, bmin;
	int rmax, gmax, bmax;
	double hmin, smin, imin;
	double hmax, smax, imax;
	
} colors[NUM_COLORS];
int curColor = 0;

#pragma mark ---- Forward declarations ----

Boolean MySGModalFilterProc (
     DialogPtr            theDialog,
     const EventRecord    *theEvent,
     short                *itemHit,
     long                 refCon );

SeqGrabComponent
MakeSequenceGrabber(WindowRef pWindow);

OSErr 
MakeSequenceGrabChannel(SeqGrabComponent seqGrab, SGChannel *sgchanVideo);

OSErr
RequestSGSettings(  SeqGrabComponent		seqGrab, 
					SGChannel				sgchanVideo);


OSErr
vdgGetSettings(VdigGrab* pVdg);

void
ImageProcessing(VdigGrab* pVdg, long dataSize);

void createNewColorSpec();

void initColorSpecs();

void printPixelInfo(int x, int y);

inline int isCurColor(int Y, int U, int V);

int findConnectedComponents(int);

void printAllColorValuesToFile(const char *filename);
void loadAllColorValuesFromFile(const char *filename);

int returnLandmark(char buf[MAXLINE]);

#pragma mark ---- VdigGrab functions ----

void vdgSetKey(unsigned char key, int modifier)
{
	//OSErr err;
	//MODIFIERS
	//GLUT_ACTIVE_SHIFT
	//GLUT_ACTIVE_CTRL
	//GLUT_ACTIVE_ALT
	switch (key)
	{
	case 'b':
	case 'B':
		createNewColorSpec();
		break;
		
	case 'g':
	case 'G':
		vdgConfigureGain(pvdg->sgchanVideo);
		break;
		
	case 'P':
		printAllColorValuesToFile("colors.txt");
		break;
		
	case 'O':
		loadAllColorValuesFromFile("colors.txt");
		break;
		
	case '-':
		if( --curColor < 0)
			curColor = 0;
		printf("Current color is %d\n", curColor);
		break;
		
	case '=':
		if( ++curColor >= NUM_COLORS)
			curColor = NUM_COLORS-1;
		printf("Current color is %d\n", curColor);
		break;
		
	case 'c':
		DRAWING_CC = !DRAWING_CC;
		break;
		
	case 'o':
		COLORING = !COLORING;
		break;
		
	default:
		break;
	}
}

void vdgSetClick(int flag, int x, int y, int modifier)
{
	switch (flag)
	{
		case 0:	//Left mouse down
			box.bottom = box.top = y;
			box.right = box.left = x;
			break;
			
		case 1:	//Left mouse up
			box.right = x;
			box.bottom = y;
			break;
			
		case 2: //Right mouse click
			printPixelInfo(x, y);
			break;
	}
}

void vdgDataProcess(char buf[MAXLINE], int *n)
{
	if(buf[0] == 'k')
	{
		printError("k-type msg arrived\n");
		//send to vdgKeySet
		vdgSetKey(buf[1], buf[2]-'0');
		printError("sending response...\n");
		sprintf(buf, "ok\n");
		*n = strlen(buf);
	}
	else if(buf[0] == 'u')
	{
		//printError("SLAM update request. Sending data...\n");
		returnLandmark(buf);
		*n = strlen(buf);
	}
	/*
	else if (buf[0] == 'C')
	{
			int cNum;
			if (sscanf(recvbuf, "C %i", &cNum) == 1)
			...
	*/
	else
	{
		sprintf(buf, "Incoming msg had size of %d\n", *n);
		*n = strlen(buf);
	}
}

VdigGrab*
vdgNew()
{
	VdigGrab* pVdg = NULL;
	pVdg = malloc(sizeof(VdigGrab));
	return pVdg;
}

OSErr
vdgInit(VdigGrab* pVdg)
{
	OSErr err;

	// zero initialize the structure
	memset(pVdg, 0, sizeof(VdigGrab));
	
	if (err = EnterMovies())
	{
		printError("EnterMovies err=%d\n", err); 
		goto endFunc;
	}

	if (!(pVdg->seqGrab = MakeSequenceGrabber(NULL)))
	{
		printError("MakeSequenceGrabber error\n"); 
		goto endFunc;
	}
	
	if (err = MakeSequenceGrabChannel(pVdg->seqGrab, &pVdg->sgchanVideo))
	{
		printError("MakeSequenceGrabChannel err=%d\n", err); 
		goto endFunc;
	}
	
	pvdg = pVdg;
	
	doSomethingCool = 0;
	
endFunc:
	return err;
}

OSErr
vdgRequestSettings(VdigGrab* pVdg)
{
	OSErr err;

	//BEN BEN BEN
	// Use the SG Dialog to allow the user to select device and compression settings
	if (err = RequestSGSettings(  pVdg->seqGrab,
								  pVdg->sgchanVideo))
	{
		printError("RequestSGSettings err=%d\n", err); 
		goto endFunc;
	}
	
	if (err = vdgGetSettings(pVdg))
	{
		printError("vdgGetSettings err=%d\n", err); 
		goto endFunc;
	}

endFunc:
	return err;
}

OSErr
vdgGetDeviceNameAndFlags(VdigGrab* pVdg, char* szName, long* pBuffSize, UInt32* pVdFlags)
{
	OSErr err;
	Str255	vdName;
    UInt32	vdFlags;

	if (!pBuffSize)
	{
		printError("vdgGetDeviceName: NULL pointer error\n");
		err = qtParamErr; 
		goto endFunc;
	}
	
	if (err = VDGetDeviceNameAndFlags(  pVdg->vdCompInst,
										vdName,
										&vdFlags))
	{
		printError("VDGetDeviceNameAndFlags err=%d\n", err);
		*pBuffSize = 0; 
		goto endFunc;
	}
	
	if (szName)
	{
		int copyLen = min(*pBuffSize-1, vdName[0]);
			
		strncpy(szName, vdName+1, copyLen);
		szName[copyLen] = '\0';

		*pBuffSize = copyLen + 1;
	} else
	{
		*pBuffSize = vdName[0] + 1;
	} 

	if (pVdFlags)
		*pVdFlags = vdFlags;

endFunc:
	return err;
}

OSErr
vdgSetDestination(  VdigGrab* pVdg,
					CGrafPtr  dstPort )
{
	pVdg->dstPort = dstPort;
	return noErr;
}

OSErr
vdgPreflightGrabbing(VdigGrab* pVdg)
{
/* from Steve Sisak (on quicktime-api list):
A much more optimal case, if you're doing it yourself is:

  VDGetDigitizerInfo() // make sure this is a compressed source only
  VDGetCompressTypes() // tells you the supported types
  VDGetMaxSourceRect() // returns full-size rectangle (sensor size)
  VDSetDigitizerRect() // determines cropping

  VDSetCompressionOnOff(true)

    VDSetFrameRate()         // set to 0 for default
    VDSetCompression()       // compresstype=0 means default
    VDGetImageDescription()  // find out image format
    VDGetDigitizerRect()     // find out if vdig is cropping for you
    VDResetCompressSequence()

    (grab frames here)

  VDSetCompressionOnOff(false)
*/
	OSErr err;
    Rect maxRect;
	
	DigitizerInfo info;
	
	// make sure this is a compressed source only
	if (err = VDGetDigitizerInfo(pVdg->vdCompInst, &info))
	{
		if (!(info.outputCapabilityFlags & digiOutDoesCompress))
		{
			printError("VDGetDigitizerInfo: not a compressed source device.\n");
			goto endFunc;
		}
	} 
	 
//  VDGetCompressTypes() // tells you the supported types

/* 
	// Apple's SoftVDig doesn't seem to like these calls
	if (err = VDCaptureStateChanging(   pVdg->vdCompInst, 
										vdFlagCaptureLowLatency | vdFlagCaptureSetSettingsBegin))
	{
		printError("VDCaptureStateChanging err=%d\n", err);
//		goto endFunc;
	}
*/

	if (err = VDGetMaxSrcRect(  pVdg->vdCompInst, currentIn, &maxRect))
	{
		printError("VDGetMaxSrcRect err=%d\n", err);
//		goto endFunc;
	}
	
	// Try to set maximum capture size ... is this necessary as we're setting the 
	// rectangle in the VDSetCompression call later?  I suppose that it is, as
	// we're setting digitization size rather than compression size here...
	// Apple vdigs don't like this call	
	if (err = VDSetDigitizerRect( pVdg->vdCompInst, &maxRect))
	{
		printError("VDSetDigitizerRect err=%d\n", err);
//		goto endFunc;		
	}

	if (err = VDSetCompressionOnOff( pVdg->vdCompInst, 1))
	{
		printError("VDSetCompressionOnOff err=%d\n", err);
//		goto endFunc;		
	}

	// We could try to force the frame rate here... necessary for ASC softvdig
	if (err = VDSetFrameRate( pVdg->vdCompInst, 0))
	{
		printError("VDSetFrameRate err=%d\n", err);
//		goto endFunc;		
	}

	// try to set a format that matches our target
	// necessary for ASC softvdig (even if it doesn't support
	// the requested codec)
	// note that for the Apple IIDC vdig in 10.3 if we request yuv2 explicitly
	// we'll get 320x240 frames returned but if we leave codecType as 0
	// we'll get 640x480 frames returned instead (which use 4:1:1 encoding on
	// the wire rather than 4:2:2)
    if (err = VDSetCompression(	pVdg->vdCompInst,
                                0, //'yuv2'
                                0,	
                                &maxRect, 
                                0, //codecNormalQuality,
                                0, //codecNormalQuality,
                                0))
	{
		printError("VDSetCompression err=%d\n", err);
//		goto endFunc;			
	}

/*
	if (err = VDCaptureStateChanging(pVdg->vdCompInst,
   									vdFlagCaptureLowLatency | vdFlagCaptureSetSettingsEnd))
	{
		printError("VDCaptureStateChanging err=%d\n", err);
//		goto endFunc;	   
	}
 */     
	if (err =  VDResetCompressSequence( pVdg->vdCompInst ))
	{
		printError("VDResetCompressSequence err=%d\n", err);
//		goto endFunc;	   
   }
 
   pVdg->vdImageDesc = (ImageDescriptionHandle)NewHandle(0);
   if (err = VDGetImageDescription( pVdg->vdCompInst, pVdg->vdImageDesc))
   {
		printError("VDResetCompressSequence err=%d\n", err);
//		goto endFunc;	   
   }

	// From Steve Sisak: find out if Digitizer is cropping for you.
	if (err = VDGetDigitizerRect( pVdg->vdCompInst, &pVdg->vdDigitizerRect))
	{
		printError("VDGetDigitizerRect err=%d\n", err);
//		goto endFunc;
	}

	pVdg->isPreflighted = 1;
	
endFunc:
	return err;
}

OSErr
vdgStartGrabbing(VdigGrab*   pVdg)
{
	OSErr err;

	if (!pVdg->isPreflighted)
	{
		printError("vdgStartGrabbing called without previous successful vdgPreflightGrabbing()\n");
		err = badCallOrderErr; 
		goto endFunc;	
	}

    if (err = VDCompressOneFrameAsync( pVdg->vdCompInst ))
	{
		printError("VDCompressOneFrameAsync err=%d\n", err);
		goto endFunc;	
	}
	
	if (err = vdgDecompressionSequenceBegin( pVdg, pVdg->dstPort, NULL, NULL ))
	{
		printError("vdgDecompressionSequenceBegin err=%d\n", err);
		goto endFunc;	
	}
	
	pVdg->isGrabbing = 1;

endFunc:
	return err;
}

OSErr
vdgGetDataRate( VdigGrab*   pVdg, 
				long*		pMilliSecPerFrame,
				Fixed*      pFramesPerSecond,
				long*       pBytesPerSecond)
{
	OSErr err;

	if (err = VDGetDataRate( pVdg->vdCompInst, 
							  pMilliSecPerFrame,
							  pFramesPerSecond,
							  pBytesPerSecond))
	{
		printError("VDGetDataRate err=%d\n", err);
		goto endFunc;		
	}
	
endFunc:	
	return err;
}

OSErr
vdgGetImageDescription( VdigGrab* pVdg,
						ImageDescriptionHandle vdImageDesc )
{
	OSErr err;
	
	if (err = VDGetImageDescription( pVdg->vdCompInst, vdImageDesc))
	{
		printError("VDGetImageDescription err=%d\n", err);
		goto endFunc;		
	}
	
endFunc:	
	return err;
}

// Make a 32bit GWorld for the decompression destination - 
// - 720 x 480 for DV. This function locks the pixel map 
OSErr MakeGWorld(short inWidth, short inHeight, 
                            GWorldPtr *outDestGWorld) 
{ 
    Rect  theBounds = {0, 0}; 
    OSErr err = noErr; 
    theBounds.right  = inWidth; 
    theBounds.bottom = inHeight; 
    *outDestGWorld = NULL; 
    err = QTNewGWorld(outDestGWorld,     // return a pointer to 
                                         // the offscreen 
                     k32ARGBPixelFormat, // the new GWorlds pixel 
                                         // format 
                     &theBounds,         // boundry and port rect 
                                         // for the offscreen PixMap 
                     NULL,               // handle to a ColorTable 
                     NULL,               // handle to a GDevice 
                     0);                 // flags 
    if (noErr == err) 
        // call LockPixels to prevent the base address for 
        // an offscreen pixel image from being moved while you 
        // draw into or copy from its pixel map 
        LockPixels(GetGWorldPixMap(*outDestGWorld)); 
    return err; 
} 


OSErr
vdgDecompressionSequenceBegin(  VdigGrab* pVdg,
								CGrafPtr dstPort, 
								Rect* pDstRect,
								MatrixRecord* pDstScaleMatrix )
{
	OSErr err;
	
// 	Rect				   sourceRect = pMungData->bounds;
//	MatrixRecord		   scaleMatrix;	

  	// !HACK! Different conversions are used for these two equivalent types
	// so we force the cType so that the more efficient path is used
	if ((*pVdg->vdImageDesc)->cType == FOUR_CHAR_CODE('yuv2'))
		(*pVdg->vdImageDesc)->cType = FOUR_CHAR_CODE('yuvu'); // kYUVUPixelFormat

	// make a scaling matrix for the sequence
//	sourceRect.right = (*pVdg->vdImageDesc)->width;
//	sourceRect.bottom = (*pVdg->vdImageDesc)->height;
//	RectMatrix(&scaleMatrix, &sourceRect, &pMungData->bounds);
			
    // begin the process of decompressing a sequence of frames
    // this is a set-up call and is only called once for the sequence - the ICM will interrogate different codecs
    // and construct a suitable decompression chain, as this is a time consuming process we don't want to do this
    // once per frame (eg. by using DecompressImage)
    // for more information see Ice Floe #8 http://developer.apple.com/quicktime/icefloe/dispatch008.html
    // the destination is specified as the GWorld
	if (err = DecompressSequenceBeginS(    &pVdg->dstImageSeq,	// pointer to field to receive unique ID for sequence
										  pVdg->vdImageDesc,	// handle to image description structure
										  0,
										  0,
										  dstPort, //pVdg->dstGWorld, //dstPort,  //GetWindowPort(pMungData->pWindow),   // port for the DESTINATION image
										  NULL,					// graphics device handle, if port is set, set to NULL
										  NULL, //&sourceRect,					// source rectangle defining the portion of the image to decompress 
                                          NULL, //&scaleMatrix,			// transformation matrix
                                          srcCopy,				// transfer mode specifier
                                          (RgnHandle)NULL,		// clipping region in dest. coordinate system to use as a mask
                                          NULL,					// flags
                                          codecHighQuality, //codecNormalQuality, 	// accuracy in decompression
                                          bestSpeedCodec)) //anyCodec); //bestSpeedCodec);		// compressor identifier or special identifiers ie. bestSpeedCodec
	{
		printError("DecompressSequenceBeginS err=%d\n", err);
		goto endFunc;
	}
		  
endFunc:	
	return err;
}

OSErr
vdgDecompressionSequenceWhen(   VdigGrab* pVdg,
								Ptr theData,
								long dataSize)
{
	OSErr err;
	CodecFlags	ignore = 0;
  
	if(err = DecompressSequenceFrameWhen(   pVdg->dstImageSeq,	// sequence ID returned by DecompressSequenceBegin
											theData,			// pointer to compressed image data
											dataSize,			// size of the buffer
											0,					// in flags
											&ignore,			// out flags
											NULL,				// async completion proc
											NULL ))
	{
		printError("DecompressSequenceFrameWhen err=%d\n", err);
		goto endFunc;
	}
	
endFunc:
	return err;
}

OSErr
vdgDecompressionSequenceEnd( VdigGrab* pVdg )
{
	OSErr err;

	if (!pVdg->dstImageSeq)
	{
		printError("vdgDestroyDecompressionSequence NULL sequence\n");
		err = qtParamErr; 
		goto endFunc;
	}

	if (err = CDSequenceEnd(pVdg->dstImageSeq))
	{
		printError("CDSequenceEnd err=%d\n", err);
		goto endFunc;
	}
	
	pVdg->dstImageSeq = 0;

endFunc:
	return err;
}

OSErr
vdgStopGrabbing(VdigGrab* pVdg)
{
	OSErr err;

	if (err = VDSetCompressionOnOff( pVdg->vdCompInst, 0))
	{
		printError("VDSetCompressionOnOff err=%d\n", err);
//		goto endFunc;		
	}

	if (err = vdgDecompressionSequenceEnd(pVdg))
	{
		printError("vdgDecompressionSequenceEnd err=%d\n", err);
//		goto endFunc;
	}

	pVdg->isGrabbing = 0;

//endFunc:
	return err;
}

bool
vdgIsGrabbing(VdigGrab* pVdg)
{
	return pVdg->isGrabbing;
}

OSErr
vdgIdle(VdigGrab* pVdg, int*  pIsUpdated)
{
	OSErr err;

    UInt8 		queuedFrameCount;
    Ptr			theData;
    long		dataSize;
    UInt8		similarity;
    TimeRecord	time;

	*pIsUpdated = 0;

	// should be while?
	if ( !(err = vdgPoll( pVdg,
						&queuedFrameCount,
						&theData,
						&dataSize,
						&similarity,
						&time))
			&& queuedFrameCount)
	{
		*pIsUpdated = 1;
		
		// Decompress the sequence
		if (err = vdgDecompressionSequenceWhen( pVdg,
												theData,
												dataSize))
		{
			printError("vdgDecompressionSequenceWhen err=%d\n", err);
//			goto endFunc;	
		}
		
		//BEGIN IMAGE PROCESSING
		ImageProcessing(pVdg, dataSize);
		// END IMAGE PROCESSING
		
		// return the buffer
		if(err = vdgReleaseBuffer(pVdg, theData))
		{
			printError("vdgReleaseBuffer err=%d\n", err);
//			goto endFunc;
		}
	}

	if (err)
	{
		printError("vdgPoll err=%d\n", err);
		goto endFunc;
	}

endFunc:
	return err;
}
	

OSErr
vdgPoll(	VdigGrab*   pVdg,
			UInt8*		pQueuedFrameCount,
			Ptr*		pTheData,
			long*		pDataSize,
			UInt8*		pSimilarity,
			TimeRecord*	pTime )
{
	OSErr err;

	if (!pVdg->isGrabbing)
	{ 
		printError("vdgGetFrame error: not grabbing\n");
		err = qtParamErr; 
		goto endFunc;
	}
	
    if (err = VDCompressDone(	pVdg->vdCompInst,
								pQueuedFrameCount,
								pTheData,
								pDataSize,
								pSimilarity,
								pTime ))
	{
		printError("vdgGetFrame error: not grabbing\n");
		goto endFunc;
	}
	
	// Overlapped grabbing
    if (*pQueuedFrameCount)
    {
		if (err = VDCompressOneFrameAsync(pVdg->vdCompInst))
		{
			printError("VDCompressOneFrameAsync err=%d\n", err);
			goto endFunc;		
		}
	}

endFunc:
	return err;
}

OSErr
vdgReleaseBuffer(VdigGrab*   pVdg, Ptr theData)
{
	OSErr err;

	if (err = VDReleaseCompressBuffer(pVdg->vdCompInst, theData))
	{
		printError("VDReleaseCompressBuffer err=%d\n", err);
		goto endFunc;		
	}

endFunc:
	return err;
}

OSErr
vdgUninit(VdigGrab* pVdg)
{
	OSErr err = noErr;		

	if (pVdg->vdImageDesc)
	{
		DisposeHandle((Handle)pVdg->vdImageDesc);
		pVdg->vdImageDesc = nil;
	}

	if (pVdg->vdCompInst)
	{
		if (err = CloseComponent(pVdg->vdCompInst))
			printError("CloseComponent err=%d\n", err);		
		pVdg->vdCompInst = nil;
	}

	if (pVdg->sgchanVideo)
	{
		if (err = SGDisposeChannel(pVdg->seqGrab, pVdg->sgchanVideo))
			printError("SGDisposeChannel err=%d\n", err);	
		pVdg->sgchanVideo = nil;
	}
	
	if (pVdg->seqGrab)
	{
		if (err = CloseComponent(pVdg->seqGrab))
			printError("CloseComponent err=%d\n", err);		
		pVdg->seqGrab = nil;
	}
	
	ExitMovies();
	return err;
}

void
vdgDelete(VdigGrab* pVdg)
{
	if (!pVdg)
	{
		printError("vdgDelete NULL pointer\n");
		return;
	}
	
	free(pVdg); 
}

#pragma mark ---- Vdig utility functions ----

OSErr
vdgGetSettings(VdigGrab* pVdg)
{	
	OSErr err;

	// Extract information from the SG
    if (err = SGGetVideoCompressor (	pVdg->sgchanVideo, 
										&pVdg->cpDepth,
										&pVdg->cpCompressor,
										&pVdg->cpSpatialQuality, 
										&pVdg->cpTemporalQuality, 
										&pVdg->cpKeyFrameRate ))
	{
		printError("SGGetVideoCompressor err=%d\n", err);
		goto endFunc;
	}
    
	if (err = SGGetFrameRate(pVdg->sgchanVideo, &pVdg->cpFrameRate))                                                                 
	{
		printError("SGGetFrameRate err=%d\n", err);
		goto endFunc;
	}
	
	// Get the selected vdig from the SG
    if (!(pVdg->vdCompInst = SGGetVideoDigitizerComponent(pVdg->sgchanVideo)))
    {
		printError("SGGetVideoDigitizerComponent error\n");
		goto endFunc;
	}
	
endFunc:
	return err;
}


// --------------------
// MakeSequenceGrabber  (adapted from Apple mung sample)
//
SeqGrabComponent
MakeSequenceGrabber(WindowRef pWindow)
{
	SeqGrabComponent seqGrab = NULL;
	OSErr			 err = noErr;

	// open the default sequence grabber
    if(!(seqGrab = OpenDefaultComponent(SeqGrabComponentType, 0)))
	{
		printError("OpenDefaultComponent failed to open the default sequence grabber.\n");
		goto endFunc;
	}
	
   	// initialize the default sequence grabber component
   	if(err = SGInitialize(seqGrab))
	{
		printError("SGInitialize err=%d\n", err);
		goto endFunc;
	}
	
	// This should be defaulted to the current port according to QT doco
	if (err = SGSetGWorld(seqGrab, GetWindowPort(pWindow), NULL))
	{
		printError("SGSetGWorld err=%d\n", err);
		goto endFunc;
	}
	
	// specify the destination data reference for a record operation
	// tell it we're not making a movie
	// if the flag seqGrabDontMakeMovie is used, the sequence grabber still calls
	// your data function, but does not write any data to the movie file
	// writeType will always be set to seqGrabWriteAppend
  	if (err = SGSetDataRef(seqGrab,
    						   0,
    						   0,
    						   seqGrabDontMakeMovie))
    {
		printError("SGSetGWorld err=%d\n", err);
		goto endFunc;
	}

endFunc:	
    if (err && (seqGrab != NULL)) 
	{ // clean up on failure
    	CloseComponent(seqGrab);
        seqGrab = NULL;
    }
    
	return seqGrab;
}


// --------------------
// MakeSequenceGrabChannel (adapted from Apple mung sample)
//
OSErr 
MakeSequenceGrabChannel(SeqGrabComponent seqGrab, SGChannel* psgchanVideo)
{
    long  flags = 0;
    OSErr err = noErr;
    
    if (err = SGNewChannel(seqGrab, VideoMediaType, psgchanVideo))
	{
		printError("SGNewChannel err=%d\n", err);
		goto endFunc;
	}
	
//	err = SGSetChannelBounds(*sgchanVideo, rect);
   	// set usage for new video channel to avoid playthrough
	// note we don't set seqGrabPlayDuringRecord
	if (err = SGSetChannelUsage(*psgchanVideo, flags | seqGrabRecord))
	{
		printError("SGSetChannelUsage err=%d\n", err);
		goto endFunc;
	}

endFunc:
    if ((err != noErr) && psgchanVideo) {
        // clean up on failure
        SGDisposeChannel(seqGrab, *psgchanVideo);
        *psgchanVideo = NULL;
    }
	
	return err;
}

OSErr
RequestSGSettings(  SeqGrabComponent		seqGrab, 
					SGChannel				sgchanVideo )
{
	OSErr err;
    	            
    SGModalFilterUPP MySGModalFilterUPP;
	if (!(MySGModalFilterUPP = NewSGModalFilterUPP (MySGModalFilterProc)))
	{
		printError("NewSGModalFilterUPP error\n");
		err = -1; // TODO appropriate error code
		goto endFunc;
	}
    
	// let the user configure and choose the device and settings
	// "Due to a bug in all versions QuickTime 6.x for the function call "SGSettingsDialog()" 
	// when used with the "seqGrabSettingsPreviewOnly" parameter, all third party panels will 
	// be excluded."  	from http://www.outcastsoft.com/ASCDFG1394.html  15/03/04
	//if (err = SGSettingsDialog(seqGrab, sgchanVideo, 0, NULL, seqGrabSettingsPreviewOnly, MySGModalFilterUPP, 0))
	if (err = SGSettingsDialog(seqGrab, sgchanVideo, 0, NULL, 0, MySGModalFilterUPP, 0))
	{
		printError("SGSettingsDialog err=%d\n", err);
		goto endFunc;
	}
    
	// Dispose of the UPP
	if (MySGModalFilterUPP)
		DisposeSGModalFilterUPP(MySGModalFilterUPP);
    
endFunc:
	return err;
}


// From QT sample code
// Declaration of a typical application-defined function
Boolean MySGModalFilterProc (
     DialogPtr            theDialog,
     const EventRecord    *theEvent,
     short                *itemHit,
     long                 refCon )
{
	// Ordinarily, if we had multiple windows we cared about, we'd handle
	// updating them in here, but since we don't, we'll just clear out
	// any update events meant for us
	Boolean	handled = false;
	
	if ((theEvent->what == updateEvt) && 
		((WindowPtr) theEvent->message == (WindowPtr) refCon))
	{
		BeginUpdate ((WindowPtr) refCon);
		EndUpdate ((WindowPtr) refCon);
		handled = true;
	}
	return (handled);
}

ComponentResult vdgConfigureGain(SGChannel inChannel)
{
    QTAtomContainer         atomContainer;
    QTAtom                  featureAtom;
    VDIIDCFeatureSettings   settings;
    VideoDigitizerComponent vd;
    ComponentDescription    desc;
    ComponentResult         result = paramErr;

    if (NULL == inChannel) goto bail;

    // get the digitizer and make sure it's legit
    vd = SGGetVideoDigitizerComponent(inChannel);
    if (NULL == vd) goto bail;

    GetComponentInfo((Component)vd, &desc, NULL, NULL, NULL);
    if (vdSubtypeIIDC != desc.componentSubType) goto bail;

    // *** now do the real work ***

    // return the gain feature in an atom container
    result = VDIIDCGetFeaturesForSpecifier(vd, vdIIDCFeatureGain, &atomContainer);
    if (noErr == result) {

        // find the feature atom
        featureAtom = QTFindChildByIndex(atomContainer, kParentAtomIsContainer,
                                         vdIIDCAtomTypeFeature, 1, NULL);
        if (0 == featureAtom) { result = cannotFindAtomErr; goto bail; }

        // find the gain settings from the feature atom and copy the data
        // into our settings
        result = QTCopyAtomDataToPtr(atomContainer,
                                     QTFindChildByID(atomContainer, featureAtom,
                                     vdIIDCAtomTypeFeatureSettings,
                                     vdIIDCAtomIDFeatureSettings, NULL),
                                     true, sizeof(settings), &settings, NULL);
        if (noErr == result) {
            /* When indicating capabilities, the flag being set indicates that the
               feature can be put into the given state.
               When indicating/setting state, the flag represents the current/desired
               state. Note that certain combinations of flags are valid for capabilities
               (i.e. vdIIDCFeatureFlagOn | vdIIDCFeatureFlagOff) but are mutually
               exclusive for state.
             */
            // is the setting supported?
            if (settings.capabilities.flags & (vdIIDCFeatureFlagOn |
                                               vdIIDCFeatureFlagManual |
                                               vdIIDCFeatureFlagRawControl)) {
                // set state flags
                settings.state.flags = (vdIIDCFeatureFlagOn |
                                        vdIIDCFeatureFlagManual |
                                        vdIIDCFeatureFlagRawControl);

                // set value - will either be 500 or the max value supported by
                // the camera represented in a float between 0 and 1.0
                settings.state.value = (1.0 / settings.capabilities.rawMaximum) *
                                       ((settings.capabilities.rawMaximum > 500) ? 500 :
                                       settings.capabilities.rawMaximum);

                // store the result back in the container
                result = QTSetAtomData(atomContainer,
                                       QTFindChildByID(atomContainer, featureAtom,
                                       vdIIDCAtomTypeFeatureSettings,
                                       vdIIDCAtomIDFeatureSettings, NULL),
                                       sizeof(settings), &settings);
                if (noErr == result) {
                    // set it on the device
                    result = VDIIDCSetFeatures(vd, atomContainer);
                }
            } else {
                // can't do it!
                result = featureUnsupported;
            }
        }
    }

bail:
    return result;
}

#pragma mark ---- Public utility functions ----

OSErr
createOffscreenGWorld(  GWorldPtr* pGWorldPtr,
						OSType		pixelFormat,
						Rect*		pBounds)
{   
	OSErr err;
	CGrafPtr theOldPort;
    GDHandle theOldDevice;
	
    // create an offscreen GWorld 
    if (err = QTNewGWorld(  pGWorldPtr,				// returned GWorld
    					pixelFormat,			// pixel format
    					pBounds,				// bounds
    					0,						// color table
    					NULL,					// GDHandle
    					0))						// flags
	{
		printError("QTNewGWorld: err=%d\n", err);
		goto endFunc;
	}
    
    // lock the pixmap and make sure it's locked because
    // we can't decompress into an unlocked PixMap
    if(!LockPixels(GetGWorldPixMap(*pGWorldPtr)))
		printError("createOffscreenGWorld: Can't lock pixels!\n");
    
    GetGWorld(&theOldPort, &theOldDevice);    
    SetGWorld(*pGWorldPtr, NULL);
    BackColor(blackColor);
    ForeColor(whiteColor);
    EraseRect(pBounds);    
    SetGWorld(theOldPort, theOldDevice);

endFunc:
	return err;
}

void
disposeOffscreenGWorld(GWorldPtr gworld)
{
	UnlockPixels(GetGWorldPixMap(gworld));
    DisposeGWorld(gworld);
}

#pragma mark ---- Vision Functions ----

/*
		UYVY means -> YUV 4:2:2 (Y sample at every pixel, U and V sampled at every second pixel horizontally on each line)
		
		R = Y + 1.402 (V-128)
		G = Y - 0.34414 (U-128) - 0.71414 (V-128)
		B = Y + 1.772 (U-128)
		
		Y = ( (  66 * R + 129 * G +  25 * B + 128) >> 8) +  16;
		U = ( ( -38 * R -  74 * G + 112 * B + 128) >> 8) + 128;
		V = ( ( 112 * R -  94 * G -  18 * B + 128) >> 8) + 128;
		http://msdn.microsoft.com/library/default.asp?url=/library/en-us/wceddraw/html/_dxce_converting_between_yuv_and_rgb.asp
		
		http://www.fourcc.org/fccyvrgb.php
*/

int DO_FRAME = 1;

/*
This is called for every frame.
*/
void ImageProcessing(VdigGrab* pVdg, long dataSize)
{
	PixMapHandle hPixMap     = GetGWorldPixMap(pVdg->dstPort); 
	long         theRowBytes = QTGetPixMapHandleRowBytes(hPixMap); 
	unsigned char* pPixels   = (unsigned char*)GetPixBaseAddr(hPixMap);
	
	unsigned char U, Y, V;
	int R=255, G=0, B=0;
	int i, count=0;
	long xtotal=0, ytotal=0;
	Y = ( (  66 * R + 129 * G +  25 * B + 128) >> 8) + 16;
	U = ( ( -38 * R -  74 * G + 112 * B + 128) >> 8) + 128;
	V = ( ( 112 * R -  94 * G -  18 * B + 128) >> 8) + 128;
	R = 0;
	G = 255;
	int Y2 = ( (  66 * R + 129 * G +  25 * B + 128) >> 8) + 16;
	int U2 = ( ( -38 * R -  74 * G + 112 * B + 128) >> 8) + 128;
	int V2 = ( ( 112 * R -  94 * G -  18 * B + 128) >> 8) + 128;
	
	if(DRAWING_CC)
	{
		if(DO_FRAME)
		{
			printf("Num components found: %d\n", findConnectedComponentsDFS(curColor));
		}
		DO_FRAME = !DO_FRAME;
	}
		
	if(COLORING)
	{
		for(i=0; i<dataSize; i+=4)
		{
			if(isCurColor(pPixels[i+1], pPixels[i], pPixels[i+2]))
			{
				pPixels[i] = U2;
				pPixels[i+1] = Y2;
				pPixels[i+2] = V2;
				pPixels[i+3] = Y2;
				xtotal += (i % theRowBytes) / 2;
				ytotal += i/theRowBytes;
				count++;
			}
		}
		//printf("Avg is: (%d, %d) from %d pixels.\n", (int)(xtotal/count), (int)(ytotal/count), (int)(count));
	}
	
	if(0)//COLORING)
	{
		// Iterative thresholding over Y
		int bin = 5;
		int N[51];	// 255/5
		int ni;
		for(int n = 0; n<51; n++)
			N[n] = 0;
		int y;
		//for(;i<dataSize/2+theRowBytes;i+=4)
		for(i=0; i<dataSize; i+=4)
		{
			y = pPixels[i+1];
			ni = Y/bin;
			N[ni]++;
			y = pPixels[i+3];
			ni = Y/bin;
			N[ni]++;
		}
		int Tk = 130;
		int Tl = 0;
		int left = 0;
		int right = 0;
		int lTot = 0;
		int rTot = 0;
		while( Tl-Tk > bin || Tl-Tk < -bin )
		{
			Tl = Tk;
			for(i = 0; i<Tk/bin; i++)
			{
				left = left + i*bin*N[i];
				lTot = lTot + N[i];
			}
			for(i = Tk/bin; i<51; i++)
			{
				right = right + i*bin*N[i];
				rTot = rTot + N[i];
			}
			Tk = (left/lTot + right/rTot)*0.5;
		}
		printf("threshold is: %d\n", Tk);
		for(i=0; i<dataSize; i+=4)
		{
			y = pPixels[i+1];
			ni = Y/bin;
			N[ni]++;
			y = pPixels[i+3];
			
			if(y < Tk)
			{
				pPixels[i] = U;
				pPixels[i+1] = Y;
				pPixels[i+2] = V;
				pPixels[i+3] = Y;
				//xtotal += (i % theRowBytes) / 2;
				//ytotal += i/theRowBytes;
				//count++;
			}
		}
	}
	
	/*
	int r[2], g[2], b[2];
	
	for(i=0;i<dataSize/4;i+=4)
	{
		U = pPixels[i];
		Y = pPixels[i+1];
		V = pPixels[i+2];
		Y2 = pPixels[i+3];
		r[0] = Y + 1.402 *(V-128);
		g[0] = Y - 0.34414 *(U-128) - 0.71414 *(V-128);
		b[0] = Y + 1.772 *(U-128);
		
		r[1] = Y2 + 1.402 *(V-128);
		g[1] = Y2 - 0.34414 *(U-128) - 0.71414 *(V-128);
		b[1] = Y2 + 1.772 *(U-128);
		
		Y = ( (  66 * R + 129 * G +  25 * B + 128) >> 8) + 16;
		U = ( ( -38 * R -  74 * G + 112 * B + 128) >> 8) + 128;
		V = ( ( 112 * R -  94 * G -  18 * B + 128) >> 8) + 128;
		
		//pPixels[i] = U;
		//pPixels[i+1] = Y;
		//pPixels[i+2] = V;
		//pPixels[i+3] = Y;
	}
	*/

}

void createNewColorSpec()
{
	//box
	PixMapHandle hPixMap     = GetGWorldPixMap(pvdg->dstPort);
	long         theRowBytes = QTGetPixMapHandleRowBytes(hPixMap);
	unsigned char* pPixels   = (unsigned char*)GetPixBaseAddr(hPixMap);
	//long		 dataSize	 = 2 * (*hPixMap)->bounds.right * (*hPixMap)->bounds.bottom;
	
	//for now we're only going to care about YUV
	unsigned char U, Y, V;
	
	int i=box.top*theRowBytes+2*box.left, x, y;
	int modulo=0;
	modulo = i%4;
	U = pPixels[i-modulo];
	Y = pPixels[i+1-modulo];
	V = pPixels[i+2-modulo];
	colors[curColor].ymin = Y;
	colors[curColor].ymax = Y;
	colors[curColor].umin = U;
	colors[curColor].umax = U;
	colors[curColor].vmin = V;
	colors[curColor].vmax = V;
	
	i=0;
	for(y=box.top; y<box.bottom; y++)
	{
		i = y*theRowBytes + 2*box.left;
		for(x=box.left; x<box.right; x+=2)
		{
			modulo = i%4;
			U = pPixels[i-modulo];
			Y = pPixels[i+1-modulo];
			V = pPixels[i+2-modulo];
			
			colors[curColor].ymin >= Y ? colors[curColor].ymin = Y : 1;
			colors[curColor].ymax <= Y ? colors[curColor].ymax = Y : 1;
			colors[curColor].umin >= U ? colors[curColor].umin = U : 1;
			colors[curColor].umax <= U ? colors[curColor].umax = U : 1;
			colors[curColor].vmin >= V ? colors[curColor].vmin = V : 1;
			colors[curColor].vmax <= V ? colors[curColor].vmax = V : 1;
			
			i += 4;
		}
	}
	
	printError("ymin =%d\n",colors[curColor].ymin);
	printError("ymax =%d\n",colors[curColor].ymax);
	printError("umin =%d\n",colors[curColor].umin);
	printError("umax =%d\n",colors[curColor].umax);
	printError("vmin =%d\n",colors[curColor].vmin);
	printError("vmax =%d\n",colors[curColor].vmax);
	
	/*unsigned char U, Y, V;
	
	int i=box.top*theRowBytes+2*box.left, x, y;
	int modulo=0;
	modulo = i%4;
	U = pPixels[i-modulo];
	Y = pPixels[i+1-modulo];
	V = pPixels[i+2-modulo];
	int r = Y + 1.402 *(V-128);
	int g = Y - 0.34414 *(U-128) - 0.71414 *(V-128);
	int b = Y + 1.772 *(U-128);
	colors[curColor].rmin = r;
	colors[curColor].rmax = r;
	colors[curColor].gmin = g;
	colors[curColor].gmax = g;
	colors[curColor].bmin = b;
	colors[curColor].bmax = b;
	
	i=0;
	
	for(y=box.top; y<box.bottom; y++)
	{
		i = y*theRowBytes + 2*box.left;
		for(x=box.left; x<box.right; x+=2)
		{
			modulo = i%4;
			U = pPixels[i-modulo];
			Y = pPixels[i+1-modulo];
			V = pPixels[i+2-modulo];
			
			r = Y + 1.402 *(V-128);
			g = Y - 0.34414 *(U-128) - 0.71414 *(V-128);
			b = Y + 1.772 *(U-128);
			
			colors[curColor].rmin >= r ? colors[curColor].rmin = r : 1;
			colors[curColor].rmax <= r ? colors[curColor].rmax = r : 1;
			colors[curColor].gmin >= g ? colors[curColor].gmin = g : 1;
			colors[curColor].gmax <= g ? colors[curColor].gmax = g : 1;
			colors[curColor].bmin >= b ? colors[curColor].bmin = b : 1;
			colors[curColor].bmax <= b ? colors[curColor].bmax = b : 1;
			
			i += 4;
		}
	}
	
	printError("rmin =%d\n",colors[curColor].rmin);
	printError("rmax =%d\n",colors[curColor].rmax);
	printError("gmin =%d\n",colors[curColor].gmin);
	printError("gmax =%d\n",colors[curColor].gmax);
	printError("bmin =%d\n",colors[curColor].bmin);
	printError("bmax =%d\n",colors[curColor].bmax);*/
}

void initColorSpecs()
{
	//set all colors to something
	int i;
	for(i=0; i<NUM_COLORS; i++)
	{
		colors[i].ymin = 0;
		colors[i].umin = 0;
		colors[i].vmin = 0;
		colors[i].ymax = 0;
		colors[i].umax = 0;
		colors[i].vmax = 0;
		colors[i].rmin = 0;
		colors[i].gmin = 0;
		colors[i].bmin = 0;
		colors[i].rmax = 0;
		colors[i].gmax = 0;
		colors[i].bmax = 0;
		colors[i].hmin = 0;
		colors[i].smin = 0;
		colors[i].imin = 0;
		colors[i].hmax = 0;
		colors[i].smax = 0;
		colors[i].imax = 0;
	}
}

void printPixelInfo(int x, int y)
{
	PixMapHandle hPixMap     = GetGWorldPixMap(pvdg->dstPort);
	long         theRowBytes = QTGetPixMapHandleRowBytes(hPixMap);
	Ptr          pPixels     = GetPixBaseAddr(hPixMap);
	
	unsigned char U, Y, V;
	
	int i = y*theRowBytes + 2*x;
	int modulo = i%4;
	U = pPixels[i-modulo];
	Y = pPixels[i+1-modulo];
	V = pPixels[i+2-modulo];
	
	//int r = Y + 1.402 *(V-128);
	//int g = Y - 0.34414 *(U-128) - 0.71414 *(V-128);
	//int b = Y + 1.772 *(U-128);
	
	printError("Y, U, V=%d, %d, %d\n",Y,U,V);
	//printError("r, g, b=%d, %d, %d\n",r,g,b);
	if(isCurColor( Y,  U,  V))
		printError("IS CUR COLOR\n");
}


//int isCurColor(unsigned char Y, unsigned char U, unsigned char V)
inline int isCurColor(int Y, int U, int V)
{
	/*int r = Y + 1.402 *(V-128);
	int g = Y - 0.34414 *(U-128) - 0.71414 *(V-128);
	int b = Y + 1.772 *(U-128);
	
	return (colors[curColor].rmin <= r &&
			colors[curColor].rmax >= r &&
			colors[curColor].gmin <= g &&
			colors[curColor].gmax >= g &&
			colors[curColor].bmin <= b &&
			colors[curColor].bmax >= b );*/
	
	return (colors[curColor].ymin <= Y &&
			colors[curColor].ymax >= Y &&
			colors[curColor].umin <= U &&
			colors[curColor].umax >= U &&
			colors[curColor].vmin <= V &&
			colors[curColor].vmax >= V );
}

inline int isObjectColor(int Y, int U, int V, int i)
{
	return (colors[i].ymin <= Y &&
			colors[i].ymax >= Y &&
			colors[i].umin <= U &&
			colors[i].umax >= U &&
			colors[i].vmin <= V &&
			colors[i].vmax >= V );
}

inline int isObjectColor2(int x, int y)
{
	return (x > 100 && x < 200 && y>100 && y<200);
}

inline int inBounds(int x, int y)
{
	return (x>=0 && y>=0 && x<640 && y<480);
}

inline int getIndex(int x, int y)
{
	return y*640 + x;
}

inline int getRealIndex(int x, int y)
{
	return y*1280 + x*2;
}

inline void getYUV(int index, Ptr pPixels, unsigned char *Y, unsigned char *U, unsigned char *V)
{
	if(index%4==0)
	{
		*U = pPixels[index];
		*V = pPixels[index+2];
	}
	else
	{
		*V = pPixels[index];
		*U = pPixels[index-2];
	}
	*Y = pPixels[index+1];
}

int N = 0;

void DFS(int ccn, int x, int y, int *seen, Ptr pPixels, struct connectedComponent *cc, int curObj)
{
	if(!inBounds(x, y))
		return;
	int i = getIndex(x, y);
	if(seen[i] == 0)
	{
		unsigned char Y, U, V;
		int ni = getRealIndex(x, y);
		getYUV(ni, pPixels, &Y, &U, &V);
		
		if(isObjectColor(Y,U,V,curObj))//isObjectColor2(x, y))
		{
			seen[i] = ccn;
			ccAddPixel(cc, x, y);
			DFS(ccn, x+1, y, seen, pPixels, cc, curObj);
			DFS(ccn, x+1, y+1, seen, pPixels, cc, curObj);
			DFS(ccn, x-1, y, seen, pPixels, cc, curObj);
			DFS(ccn, x-1, y-1, seen, pPixels, cc, curObj);
			DFS(ccn, x, y+1, seen, pPixels, cc, curObj);
			DFS(ccn, x, y-1, seen, pPixels, cc, curObj);
			DFS(ccn, x-1, y+1, seen, pPixels, cc, curObj);
			DFS(ccn, x+1, y-1, seen, pPixels, cc, curObj);
		}
		else
			seen[i] = -1;
	}
}

inline inBoundsI(int i)
{
	return (i>=0 && i<307200);
}

int seen[307200];

int table[10000];
int tlen;

inline int checkAndSet(int ccn, int i, int *seen, Ptr pPixels, int curObj)
{
	unsigned char Y, U, V;
	getYUV(i, pPixels, &Y, &U, &V);
	if(isObjectColor(Y, U, V, curObj))
	{
		int left = 0;
		if(inBoundsI(i/2-1))
			if(seen[i/2-1] > 0)
				left = seen[i/2-1];
		int ni = (i/2)-640;
		int leftup = 0, rightup = 0, up = 0;
		if(inBoundsI(ni-1))
			if(seen[ni-1] > 0)
				if(ni%640 != 0)
					leftup = seen[ni-1];
		if(inBoundsI(ni))
			if(seen[ni] > 0)
				up = seen[ni];
		if(inBoundsI(ni+1))
			if(seen[ni+1] > 0)
				if(ni%640 != 639)
					rightup = seen[ni+1];
		
		if(left > 0)
		{
			ccn = left;
		}
		if(rightup != ccn && rightup > 0)
		{
			if(left == 0)
				ccn = rightup;
			else
			{
				table[tlen++] = ccn;
				table[tlen++] = rightup;
			}
		}
		if(leftup != ccn && leftup != rightup && leftup > 0)
		{
			if(left == 0 && rightup == 0)
				ccn = leftup;
			else
			{
				table[tlen++] = ccn;
				table[tlen++] = leftup;
			}
		}
		if(up != ccn && up != rightup && up != leftup && up > 0)
		{
			if(left == 0 && rightup == 0 && leftup == 0)
				ccn = up;
			else
			{
				table[tlen++] = ccn;
				table[tlen++] = up;
			}
		}
		
		if(left == 0 && rightup == 0 && leftup == 0 && up == 0)
			ccn++;
		
		seen[i/2] = ccn;
		
		return ccn;
	}
	else
	{
		seen[i/2] = -1;
		return ccn;
	}
}

int eqiv(int c, int i)
{
	//table[10000];
	//tlen;
	
	for(int j = 0; j<tlen; j++)
	{
		if(table[j] == i)
		{
			if(c == table[j + 1 - (2 * (j%2))])
				return 1;
		}
	}
	
	return 0;
}

void moreProcessing(int ccn)
{
	int realCompNum = 0;
	
	for(int i = 2; i<3/*ccn*/ ; i++)
	{
		ccClear(&cc);
		for(int x = 0; x<640; x++)
			for(int y = 0; y<480; y++)
				if(seen[getIndex(x, y)] == i || eqiv(seen[getIndex(x, y)], i))
					ccAddPixel(&cc, x, y);
			
		if(cc.nextEL > 300)
		{
			int xavg = 0, yavg = 0;
			for(int j=0;j<cc.nextEL;j++)
			{
				xavg += cc.x[j];
				yavg += cc.y[j];
			}
			xavg /= cc.nextEL;
			yavg /= cc.nextEL;
			initComponent(&(components[realCompNum]), cc.nextEL, xavg, yavg, curColor, getInfo(cc.x, cc.y, cc.nextEL));
			realCompNum++;
		}
	}
	
	printf("found %d real components.\n", realCompNum);
	
}

int findConnectedComponents(int curObj)
{
	PixMapHandle hPixMap     = GetGWorldPixMap(pvdg->dstPort);
	long         theRowBytes = QTGetPixMapHandleRowBytes(hPixMap);
	Ptr          pPixels     = GetPixBaseAddr(hPixMap);
	int size = (theRowBytes/2)*480;
	for(int i = 0; i < size; ++i)
		seen[i] = 0;
		
	int ccn = 1;
	tlen = 0;
		
	long dataSize = theRowBytes*480;
	for(int i = 0; i<dataSize; i+=2)
	{
		ccn = checkAndSet(ccn, i, seen, pPixels, curObj);
	}
	
	moreProcessing(ccn);
	
	if(DRAWING_CC)
	{
		unsigned char U, Y, V;
		int R=0, G=0, B=255;
		Y = ( (  66 * R + 129 * G +  25 * B + 128) >> 8) + 16;
		U = ( ( -38 * R -  74 * G + 112 * B + 128) >> 8) + 128;
		V = ( ( 112 * R -  94 * G -  18 * B + 128) >> 8) + 128;
		
		for(int i=0; i<dataSize; i+=2)
		{
			if(seen[i/2] > 0)
			{
				pPixels[i] = i%4==0? U : V;
				pPixels[i+1] = Y;
				//pPixels[i+2] = i%4==0? V : U;
				//pPixels[i+3] = Y;
			}
		}
	}
	
	return ccn-1;
}

int findConnectedComponentsDFS(int curObj)
{
	PixMapHandle hPixMap     = GetGWorldPixMap(pvdg->dstPort);
	long         theRowBytes = QTGetPixMapHandleRowBytes(hPixMap);
	Ptr          pPixels     = GetPixBaseAddr(hPixMap);
	int size = (theRowBytes/2)*480;
	//char* seen = (char *) malloc(sizeof(char)*size); // was originally an int
	for(int i = 0; i < size; ++i)
		seen[i] = 0;
		
	int ccn = 1;
	
	//ccInit(&cc);
	int x=0, y=0;
	long dataSize = theRowBytes*480;
	for(int i = 0; i<dataSize; i+=2)
	{
		if(seen[i/2] == 0)
		{
			ccClear(&cc);
			x = (i%theRowBytes)/2;
			y = i/theRowBytes;
			DFS(ccn, x, y, seen, pPixels, &cc, curObj);
			if(cc.nextEL > 300)
			{
				int xavg = 0, yavg = 0;
				for(int j=0;j<cc.nextEL;j++)
				{
					xavg += cc.x[j];
					yavg += cc.y[j];
				}
				xavg /= cc.nextEL;
				yavg /= cc.nextEL;
				initComponent(&(components[ccn - 1]), cc.nextEL, xavg, yavg, curObj, getInfo(cc.x, cc.y, cc.nextEL));
				ccn++;
			}
		}
	}
	
	if(DRAWING_CC)
	{
		unsigned char U, Y, V;
		int R=0, G=0, B=255;
		Y = ( (  66 * R + 129 * G +  25 * B + 128) >> 8) + 16;
		U = ( ( -38 * R -  74 * G + 112 * B + 128) >> 8) + 128;
		V = ( ( 112 * R -  94 * G -  18 * B + 128) >> 8) + 128;
		
		for(int i=0; i<dataSize; i+=2)
		{
			if(seen[i/2] == 1)
			{
				pPixels[i] = i%4==0? U : V;
				pPixels[i+1] = Y;
			}
		}
	}
	
	N = ccn-1;
	return ccn-1;
}

int findConnectedComponentsCV(int curObj)
{
	struct list Q;
	Q.front = Q.last = NULL;
	
	//list<CvPoint> Q;
	//uchar *color;
	unsigned char Y, U, V;
	
	PixMapHandle hPixMap     = GetGWorldPixMap(pvdg->dstPort);
	long         theRowBytes = QTGetPixMapHandleRowBytes(hPixMap);
	Ptr          pPixels     = GetPixBaseAddr(hPixMap);
	
	int xlow = 0;
	int xhi = theRowBytes/2;	//640
	int ylow = 0;
	int yhi = 480;
	int xdiff = xhi - xlow;
	int ydiff = yhi - ylow;

	int size = (xhi - xlow) * (yhi - ylow);
	//seen[size]
	char* seen = (char *) malloc(sizeof(char)*size); // was originally an int
	for(int i = 0; i < size; ++i)
		seen[i] = 0;

	int ccn = 0;

	//double h, s, v;

	int pixelCount = 0;
	int totalX = 0;
	int totalY = 0;

	struct point currentPixel;

	//vector<double> xcoords, ycoords;
	struct list coords;
	coords.front = coords.last = NULL;

	int n = 3;	// Number of search 'rings'
	
	int index = 0;

	for(int x = 0; x < xdiff; ++x)
	{
		for(int y = 0; y < ydiff; ++y)
		{
			if (seen[x*(ydiff)+y] == 0)
			{
				index = x*(ydiff*2)+y;
				U = pPixels[index-index%4];
				Y = pPixels[index-index%4+1];
				V = pPixels[index-index%4+2];
				pPixels[index-index%4] = 0;
				pPixels[index-index%4+1] = 0;
				//pPixels[index-index%4+2] = 0;
				//pPixels[index-index%4+3] = 0;
				//color = cvPtr2D(frame,y+ylow,x+xlow); 
				//RGBtoHSV(color[2], color[1], color[0], &h, &s, &v);
				if (isCurColor(Y, U, V))//isObjectColor(color[2], color[1], color[0], h, s, v, curObj))
				{
					
					totalX = totalY = pixelCount = 0;
					struct point pt;
					pt.i = index;
					push_back(&Q, pt);
					while(!empty(&Q))
					{
						currentPixel = (*front(&Q));
						int curpixX = currentPixel.i/theRowBytes;
						int curpixY = (currentPixel.i%theRowBytes)/2;
						seen[(ydiff)*curpixX+curpixY] = 1;
						++pixelCount;
						totalX += (currentPixel.i%theRowBytes)/2;
						totalY += currentPixel.i/theRowBytes;
						for(int i = -n; i <= n; ++i)
						{
							for(int j = 0; j <= n; ++j)
							{
								if(curpixX + i >= 0 && curpixY + j >= 0 && 
									curpixX + i < xdiff && 
									curpixY + j < ydiff)
								{
									index = x*(theRowBytes)+y;
									U = pPixels[index-index%4];
									Y = pPixels[index-index%4+1];
									V = pPixels[index-index%4+2];
									//color = cvPtr2D(frame,currentPixel.y+j+ylow,currentPixel.x+i+xlow); 
									//RGBtoHSV(color[2], color[1], color[0], &h, &s, &v);
									if (seen[(ydiff)*(curpixX+i) + curpixY+j] == 0 && 
										isObjectColor(Y, U, V, curObj))
									{
										struct point pt;
										pt.i = (curpixX+i)*(2*ydiff)+curpixY+j;
										
										push_back(&Q, pt);
										// These draw up to three different colors for the connected components
										//...but doesn't stay drawn (because function is called from ThreadFunc)
										//uchar* painted;
										//painted = cvPtr2D(connected_comps, currentPixel.y+j+ylow, currentPixel.x+i+xlow);
										//painted[curObj <= 3 ? curObj - 1: ccn] = 255;
										pPixels[pt.i] = 240;
									}
									seen[(ydiff)*(curpixX + i) + curpixY + j] = 1;
									
								}
							}
						}
						/*xcoords.push_back(currentPixel.x+xlow);
						ycoords.push_back(currentPixel.y+ylow);
						
						struct point p;
						p.i = 
						push_back(coords, p);*/
						pop_front(&Q);
						
					}
					if (pixelCount > 70)	//a somewhat significant blob
					{
						/*struct Component newObj;
						initComponent(&newObj, pixelCount, totalX / pixelCount, totalY / pixelCount,
										curObj, getInfo(xcoords, ycoords, pixelCount));
						//newObj.maxDiameter(frame);
						connectedComponents.push_back(newObj);
						if (ccn < 2)
							ccn++;*/
						ccn++;
					}
					clear(&coords);
				}
			}
			seen[x*(ydiff)+y] = 1;
		}
	}

	/*delete[] seen;

	return (int)connectedComponents.size(); // The (int) gets rid of a warning
			// Presumably, we have few enough components to use a signed int :-)
			*/
	free(seen);
	return ccn;
}
//*/

void loadColorValues(int coi, FILE *fp)
{
	int i;
	fscanf(fp, "Color thresholds for object #%d", &i);
	if( i != coi )
		printf("Might be a reading error...%d\n", i);
	
	fscanf(fp, "  Y    %d %d\n", &(colors[coi].ymin), &(colors[coi].ymax));
	fscanf(fp, "  U    %d %d\n", &(colors[coi].umin), &(colors[coi].umax));
	fscanf(fp, "  V    %d %d\n", &(colors[coi].vmin), &(colors[coi].vmax));
}

void printColorValues(int coi, FILE *fp)
{
	fprintf(fp, "Color thresholds for object #%d\n", coi);
	fprintf(fp, "  Y    %d %d\n", colors[coi].ymin, colors[coi].ymax);
	fprintf(fp, "  U    %d %d\n", colors[coi].umin, colors[coi].umax);
	fprintf(fp, "  V    %d %d\n", colors[coi].vmin, colors[coi].vmax);
}

void printAllColorValuesToFile(const char *filename)
{
	FILE *fp = fopen(filename, "w");
	if(fp == NULL)
	{
		printf("error creating/opening file\n");
		return;
	}

	printf("** Saving color values to file ");
	printf(filename);
	printf("\n");

	// output the number of objects
	fprintf(fp, "%d\n\n", NUM_COLORS);

	for (int i=0 ; i<NUM_COLORS ; ++i)
	{
        printColorValues(i,fp);
	}

	fprintf(fp, "\n\n");

	printf("** Finished saving color values to file ");
	printf(filename);
	printf("\n");
	
	if(fclose(fp) != 0)
		printf("Error closing file.\n");
}


void loadAllColorValuesFromFile(const char *filename)
{
	FILE *fp = fopen(filename, "r");
	if(fp == NULL)
	{
		printf("error creating/opening file\n");
		return;
	}

	printf("** Loading color values from file ");
	printf(filename);
	printf("\n");

	// get the number of objects
	int numobj;
	fscanf(fp, "%d\n\n", &numobj);
	//we probably don't need to do this...
	#define NUM_COLORS numobj

	// get each object, in turn
	for (int i=0 ; i<NUM_COLORS ; ++i)
	{
        loadColorValues(i,fp);
	}

	printf("** Finished loading color values from file ");
	printf(filename);
	printf("\n");
	
	if(fclose(fp) != 0)
		printf("Error closing file.\n");
}

int blobLandmarkDetect(char buf[MAXLINE])
{
	PixMapHandle hPixMap     = GetGWorldPixMap(pvdg->dstPort); 
	long         theRowBytes = QTGetPixMapHandleRowBytes(hPixMap); 
	unsigned char* pPixels   = (unsigned char*)GetPixBaseAddr(hPixMap);
	
	const int NUM_LANDMARKS = 5;
	
	int pCount[NUM_LANDMARKS];
	int x[NUM_LANDMARKS];
	int y[NUM_LANDMARKS];
	for(int i = 0; i<NUM_LANDMARKS; i++)
	{
		pCount[i] = 0;
		x[i] = 0;
		y[i] = 0;
	}
	
	int dataSize = theRowBytes*480;
	for(int i=0; i<dataSize; i+=4)
	{
		for(int j=0; j<NUM_LANDMARKS; j++)
		{
			if(isObjectColor(pPixels[i+1], pPixels[i], pPixels[i+2], j))
			{
				x[j] += (i % theRowBytes) / 2;
				y[j] += i/theRowBytes;
				pCount[j] += 1;
			}
		}
	}
	
	int n = -1;
	int count = 50;
	for(int i = 0; i<NUM_LANDMARKS; i++)
	{
		if(pCount[i] > count)
		{
			count = pCount[i];
			n = i;
		}
	}
	
	if(n != -1)
	{
		sprintf(buf, "%d %d %d\n", n, x[n]/pCount[n], y[n]/pCount[n]);
		return 1;
	}
	
	sprintf(buf, "Err\n");
	return 0;
}

int returnLandmark(char buf[MAXLINE])
{
	//PixMapHandle hPixMap     = GetGWorldPixMap(pvdg->dstPort); 
	//long         theRowBytes = QTGetPixMapHandleRowBytes(hPixMap); 
	//unsigned char* pPixels   = (unsigned char*)GetPixBaseAddr(hPixMap);
	
	//int flag1 = DRAWING_CC;
	//int flag2 = COLORING;
	//DRAWING_CC = COLORING = 0;
	
	//zero components
	//for(int i = 0; i<100; i++)
	//{
	//	components[i].pixelCount = 0;
	//}
	
	//int n = findConnectedComponentsDFS(curColor);
	
	return blobLandmarkDetect(buf);
	
	int biggestObj = -1;
	int pCount = 0;
	for(int i = 0; i<100; i++)
	{
		if(components[i].pixelCount > pCount)
		{
			pCount = components[i].pixelCount;
			biggestObj = i;
		}
	}
	
	if(biggestObj != -1)
	{
		sprintf(buf, "%d %d %d\n", curColor, components[biggestObj].avgX, components[biggestObj].avgY);
		return 1;
	}
	
	sprintf(buf, "Err\n");
	return 0;
}






